"""Tools for managing the global config."""

from __future__ import annotations

import asyncio
import tomllib
from http import HTTPStatus
from pathlib import Path
from typing import TYPE_CHECKING

import logistro
import platformdirs
import tomli_w
import typer
from fastapi.responses import JSONResponse

from tetsuya._globals import app, cli, service_types
from tetsuya.app.client import get_client

if TYPE_CHECKING:
    from typing import Any

_logger = logistro.getLogger(__name__)

config_file = (
    Path(platformdirs.user_config_dir("tetsuya", "pikulgroup")) / "config.toml"
)

config_data: dict[Any, Any] = {}


def _load_config() -> bool:
    # need a reload
    if config_file.is_file():
        with config_file.open("rb") as f:
            config_data.clear()
            config_data.update(tomllib.load(f))
        return True
    else:
        _logger.info("No config file found.")
        return False


_load_config()

config_cli = typer.Typer(help="Manage the config.")
cli.add_typer(config_cli, name="config")


# should be able to load up *specific* log items and not replace others
# should confirm on force


# this should maybe have print? overwrite doesn't really make sense
@config_cli.command()
def touch(*, default: bool = False, force: bool = False, dump: bool = False):
    """Create config file if it doesn't exist."""
    client = get_client()
    _logger.debug("Sending touch command.")
    r = client.put(
        "/config/touch",
        json={"default": default, "force": force},
    )
    _logger.debug("Processing touch response")
    # check return value
    if r.status_code == HTTPStatus.OK:
        result = r.json()
        if not dump:
            print(result.get("path", f"Weird result: {result}"))  # noqa: T201
        else:
            print(result.get("content", f"Weird result: {result}"))  # noqa: T201
    else:
        raise ValueError(f"{r.status_code}: {r.text}")


@app.put("/config/touch")
async def _touch(data: dict):
    """Create the config file if it doesn't exist."""
    _logger.info("Touching config file.")
    _logger.info(f"Touch received data: {data}")

    config_file.parent.mkdir(parents=True, exist_ok=True)

    if data.get("default"):
        config_dict: dict[Any, Any] = {}
        for t in service_types:
            if hasattr(t, "default_config"):
                d = config_dict.setdefault(t.__name__, {})
                d.update(t.default_config())
        if not config_file.exists() or data.get("force"):
            await asyncio.to_thread(config_file.write_text, tomli_w.dumps(config_dict))
        else:
            return JSONResponse(
                content={"error": "File exists. Use --force."},
                status_code=409,
            )
    else:
        config_file.touch()
    text = config_file.read_text(encoding="utf-8")
    ret = {"path": str(config_file.resolve()), "content": text}
    _logger.info(f"Touch sending back: {ret}")
    return ret


@config_cli.command()
def reload():
    """Reload config file."""
    client = get_client()
    _logger.info("Sending reload command.")
    r = client.post(
        "/config/reload",
    )
    # check return value
    if r.status_code == HTTPStatus.OK:
        print("OK")  # noqa: T201
    else:
        raise ValueError(f"{r.status_code}: {r.text}")


@app.post("/config/reload")
async def _reload():
    """Reload config file."""
    _logger.info("Reloading config file.")
    res = _load_config()
    if not res:
        return JSONResponse(
            content={},
            status_code=404,
        )
    return None



"""Utitilies for everyone. No deps."""

from pathlib import Path

import logistro
import platformdirs

_logger = logistro.getLogger(__name__)

# The folder where we'll create a socket
runtime_dir = platformdirs.user_runtime_dir("tetsuya", "pikulgroup")


def uds_path() -> Path:
    """Return default socket path."""
    base = Path(runtime_dir)
    p = base / "tetsuya.sock"
    p.parent.mkdir(parents=True, exist_ok=True)
    _logger.info(f"Socket path: {p!s}")
    return p



"""Exposes a SearchGit service to find git repos below your home."""

import subprocess
from dataclasses import dataclass, field
from datetime import UTC, datetime, timedelta
from pathlib import Path

import logistro

from tetsuya._globals import service_types

from . import _protocol
from .utils.config import config_data

_logger = logistro.getLogger(__name__)

# These will have to be versioned
# And they will have to have migratory functions
# They should also be logged (independently of other logs)
# And their constructors should reasonably take an old version
# config validation/default maybe dataclass or typed dict
# Also should be able to consume javascript and return report


@dataclass()
class SearchGitReport:
    """Report format for SearchGit."""

    retval: int
    stderr: str
    repos: list[Path]
    created_at: datetime = field(
        default_factory=lambda: datetime.now(tz=UTC),
    )

    def long(self) -> str:
        """One full path per line."""
        return "\n".join(str(p.resolve()) for p in sorted(self.repos))

    def short(self) -> str:
        """Comma-separated last directory names."""
        return ", ".join(p.name for p in sorted(self.repos))


class SearchGit(_protocol.Bannin):  # is Bannin
    """SearchGit is a class to find git repos below your home directory."""

    report_type: type[_protocol.Output] = SearchGitReport

    @classmethod  # make mandatory through protocol
    def default_config(cls) -> dict:
        """Return dictionary with default config."""
        return {
            "ignore_folders": [".cache"],
            "ignore_paths": [],
        }

    def __init__(self):
        """Construct a SearchGit service."""
        self.name = self.__class__.__name__
        self.cachelife = timedelta(hours=12)
        self.version = 0
        self.cache = None
        # check if reloading (cache)

    def _execute(self) -> SearchGitReport:
        """Execute search of your home repository for git repos."""
        home = Path.home()

        _c = config_data.get(self.name, {})

        ignore_folders = _c.get("ignore_folders", [])
        _logger.info(f"Ignoring folders: {ignore_folders}")

        ignore_paths = _c.get("ignore_paths", [])
        _logger.info(f"Ignoring paths: {ignore_paths}")

        # Build the prune expression:
        # ( -path <abs> -o -path <abs> -o -name <nm> -o ... )
        expr = []
        for p in ignore_paths:
            expr += ["-path", str(p)]
            expr += ["-o"]
        for name in ignore_folders:
            expr += ["-name", str(name)]
            expr += ["-o"]
        # drop last -o, close group, then -prune -o
        expr = [r"(", *expr[:-1], r")", "-prune", "-o"] if expr else []

        cmd = [
            "find",
            str(home),
            *expr,
            "-type",
            "d",
            "-name",
            ".git",
            "-printf",
            r"%h\n",  # print the parent dir of .git
            "-prune",  # and don't descend into the .git dir itself
        ]
        _logger.info(" ".join(cmd))
        p = subprocess.run(  # noqa: S603
            cmd,
            check=False,
            capture_output=True,
        )
        retval, stdout, stderr = p.returncode, p.stdout, p.stderr
        _logger.info("Git search ran find process.")
        _repos = sorted(
            {Path(p) for p in stdout.decode(errors="ignore").split("\n") if p},
        )

        return SearchGitReport(retval=retval, stderr=stderr.decode(), repos=_repos)


service_types.append(SearchGit)



"""services contains the services plus their utilities."""

from __future__ import annotations

import asyncio
from dataclasses import asdict, is_dataclass
from http import HTTPStatus
from typing import TYPE_CHECKING, Literal

import logistro
import typer
from fastapi.responses import JSONResponse

from tetsuya._globals import active_services, app, cli
from tetsuya.app.client import get_client

from . import utils as utils
from .search_git import SearchGit

if TYPE_CHECKING:
    from typing import Any

__all__ = ["SearchGit"]

_logger = logistro.getLogger(__name__)

service_cli = typer.Typer(help="Manage the services.")
cli.add_typer(service_cli, name="service")


@service_cli.command(name="list")  # accept --all
def _list():
    """List running services, or all services with --all."""
    client = get_client()
    _logger.info("Sending list command.")
    r = client.post(
        "/service/list",
    )
    # check return value
    if r.status_code == HTTPStatus.OK:
        for s in r.json():
            print(s)  # noqa: T201
    else:
        raise ValueError(f"{r.status_code}: {r.text}")


@app.post("/service/list")
async def __list():
    """List running services, or all services with --all."""
    return list(active_services.keys())


@service_cli.command(name="run")  # accept --all
def run(
    name: str | None = None,
    *,
    all: bool = False,  # noqa: A002
    force: bool = False,
    format: Literal["short", "long", "json"] = "json",  # noqa: A002
    # accept a cache controller here
):
    """Run a or all services."""
    client = get_client(timeout=10)
    _logger.info("Sending run command.")
    r = client.post(
        "/service/run",
        json={
            "name": name,
            "all": all,
            "force": force,
            "format": format,
        },
    )
    # check return value
    if r.status_code == HTTPStatus.OK:
        if format == "json":
            print(r.text)  # noqa: T201
        else:
            json = r.json()
            if len(json) == 1 and name in json:
                print(json[name])  # noqa: T201
            else:
                for k, v in json.items():
                    print(f"{k}:")  # noqa: T201
                    print(v)  # noqa: T201

    else:
        raise ValueError(f"{r.status_code}: {r.text}")


@app.post("/service/run")
async def _run(data: dict):  # noqa: C901, PLR0912
    """Run a or all services."""
    _n = data.get("name")
    _logger.info(f"Received service run request: {data}")
    if not _n and not data.get("all"):
        return JSONResponse(
            content={"error": "Supply either name or --all, not both."},
            status_code=400,
        )
    if _n and data.get("all"):
        return JSONResponse(
            content={"error": "Use either name or --all, not both."},
            status_code=400,
        )
    elif data.get("force") and data.get("cache"):
        return JSONResponse(
            content={"error": "Use either --force or --cache, not both."},
            status_code=400,
        )
    if _n and _n not in active_services:
        return JSONResponse(
            content={"error": f"Service {_n} not found."},
            status_code=404,
        )
    services = {_n: active_services[_n]} if _n else active_services
    tasks = {}
    k: Any
    v: Any
    for k, v in services.items():
        tasks[k] = asyncio.create_task(
            v.run(force=data.get("force", False)),
        )
    results = {}
    for k, v in tasks.items():
        await v
        _r = services[k].get_object()
        if _r is None:
            raise RuntimeError(f"Run failed for {k}")
        match data.get("format"):
            case "short":
                results[k] = _r.short()
            case "long":
                results[k] = _r.long()
            case "json":
                if not is_dataclass(_r):
                    raise RuntimeError(f"Cache for {k} is bad.")
                else:
                    results[k] = asdict(_r)
            case _:
                _logger.error(f"Unknown format: {data.get('format')}")
    _logger.debug2(f"Returning results: {results}")
    return results



"""Bottom dependency."""

from __future__ import annotations

import asyncio
from datetime import UTC, datetime, timedelta
from typing import Protocol

import logistro

_logger = logistro.getLogger(__name__)

# I'm not entirely sure about the timer
# And it doesn't actually cache and load
# And there is no way to change it.


class Output(Protocol):
    """The object a service stores or returns."""

    created_at: datetime

    def long(self) -> str: ...
    def short(self) -> str: ...


class Bannin(Protocol):
    """The abstract idea of a service."""

    report_type: type[Output]
    name: str
    cachelife: timedelta
    version: int
    cache: Output | None
    _pending_task: asyncio.Task | None = None

    @classmethod
    def default_config(cls) -> dict: ...
    def _execute(self) -> Output: ...

    def _is_cache_live(self):
        return self.cache is not None and (
            self.cache.created_at + self.cachelife > datetime.now(tz=UTC)
        )

    def get_object(self) -> Output:
        """Get the actual latest result object."""
        if not self.cache:
            raise RuntimeError(f"Cache for {self.name} wont populate.")
        return self.cache

    def _clear_task(self, p: asyncio.Task):
        if p.cancelled():
            return
        if e := p.exception():
            _logger.error(e)  # TODO(AJP): check about _logger and exceptions
            # search exc_info in everything
            # also, this won't restart

    async def _run_in_a_while(self, time: int):
        await asyncio.sleep(time)
        await self.run()

    async def run(self, *, force=False):
        """Run the service in a cache-aware manner."""
        if not force and self._is_cache_live():
            _logger.info("Not rerunning- cache is live.")
            return
        if _p := self._pending_task:
            _p.cancel()
            await asyncio.sleep(0)  # let it cancel
        self.cache = await asyncio.to_thread(self._execute)
        _logger.debug2(f"New cache: {self.cache}")
        self._pending_task = asyncio.create_task(
            self._run_in_a_while(int(self.cachelife.total_seconds())),
        )
        self._pending_task.add_done_callback(self._clear_task)
